﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.WindowsMobile.Gestures;

namespace SmartPDA.Windows.Forms
{
	/// <summary>
	/// 指定したScrollableControlインスタンスとその子コントロールをタッチ操作でスクロールできるようにするクラス。
	/// </summary>
	/// <example> 
	/// <code>
	/// Form f = new Form();
	/// f.AutoScroll = true;
	/// Button button1 = new Button(); 
	/// CheckBox checkbox1 = new CheckBox();
	/// ListBox listBox1 = new ListBox();
	/// button1.Parent = f;
	/// checkbox1.Parent = f;
	/// listBox1.Parent = f;
	/// 
	///	this.scroll = new NullScroll();
	/// this.scroll.Excludes.Add(this.checkBox1);
	///	this.scroll.Setup(this);
	/// </code>
	/// </example>
	public class NullScroll
	{

		#region プロパティ

		/// <summary>
		/// スクロール対象コントロールを取得または設定します。
		/// </summary>
		public ScrollableControl Target;

		/// <summary>
		/// スクロール対象にしないコントロールのリストを取得または設定します。
		/// </summary>
		public List<Control> Excludes;

		/// <summary>
		/// スクロール対象にしないタイプのリストを取得または設定します。
		/// </summary>
		public List<Type> ExcludeTypes;

		#endregion

		#region プロパティ(内部)

		/// <summary>
		/// 現在マウスが押されている状態であるかどうかを示す値を取得または設定します。
		/// </summary>
		bool NowMouseDown;

		/// <summary>
		/// マウスが押された時の、マウス座標(スクリーン座標)を取得または設定します。
		/// </summary>
		Point MouseDownLocation;

		/// <summary>
		/// マウスが押された時のスクロール対象コントロールのAutoScrollPositionプロパティを取得または設定します。
		/// </summary>
		Point MouseDownScrollPosition;

		/// <summary>
		/// コントロールのハンドルをキー、そのコントロールを値としたディクショナリを取得または設定します。
		/// </summary>
		Dictionary<IntPtr, Control> ControlDictionary;

		/// <summary>
		/// WM6.5以降のジェスチャAPI を使用してスクロールを行うクラス
		/// </summary>
		AutoGestureContext agc;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// NullScrollクラスのインスタンスを初期化します。
		/// </summary>
		public NullScroll()
		{
			this.Excludes = new List<Control>();
			this.ExcludeTypes = new List<Type>();
			this.ControlDictionary = new Dictionary<IntPtr, Control>();
		}

		#endregion

		#region メソッド

		/// <summary>
		/// 指定したスクロール対象コントロールとその子コントロールを自動スクロールするようにセットアップします。
		/// </summary>
		/// <param name="control"></param>
		public void Setup(ScrollableControl control)
		{
			// ターゲットを設定する
			this.Target = control;

			// WM 6.5以降？
			if (this.IsOverWM65())
			{
				this.agc = AutoGestureContext.GetContext(this.Target);
				this.agc.IgnorePan = false;
				this.agc.IgnoreScroll = false;
				this.agc.IsHorizontallyScrollable = true;
				this.agc.IsVerticallyScrollable = true;
			}
			else
			{
				// イベントをバインドする
				this.Target.MouseDown += this.ChildControl_MouseDown;
				this.Target.MouseUp += this.ChildControl_MouseUp;
				this.Target.MouseMove += this.ChildControl_MouseMove;

				// スクロール対象コントロールの子コントロールにイベントをバインドする
				this.EventBind(this.Target, true);
			}
		}

		/// <summary>
		/// 指定したスクロール対象コントロールとその子コントロールに設定された自動スクロールを解除します。
		/// </summary>
		/// <param name="control"></param>
		public void Release()
		{
			// ターゲットが設定されていない→戻る
			if (this.Target == null)
				return;
			
			// WM 6.5以降？
			if (this.IsOverWM65())
			{
				this.agc.IgnorePan = true;
				this.agc.IgnoreScroll = true;
				this.agc.IsHorizontallyScrollable = false;
				this.agc.IsVerticallyScrollable = false;
				this.agc.Dispose();
				this.agc = null;
			}
			else
			{
				// イベントをアンバインドする
				this.Target.MouseDown -= this.ChildControl_MouseDown;
				this.Target.MouseUp -= this.ChildControl_MouseUp;
				this.Target.MouseMove -= this.ChildControl_MouseMove;

				// スクロール対象コントロールの子コントロールにイベントをバインドする
				this.EventBind(this.Target, false);
			}
		}

		public void Dispose()
		{
			this.Release();
		}

		/// <summary>
		/// 指定したコントロールとその子コントロールに対して再帰的にスクロールイベントをバインドします。
		/// </summary>
		/// <param name="control"></param>
		private void EventBind(Control control, bool bind)
		{
			// スクロール対象コントロールがない→戻る
			if (control == null)
				return;

			// 子コントロールにイベントをバインドする
			foreach (Control child in control.Controls)
			{
				// 除外リストに含まれている→次へ
				if (this.Excludes != null && this.Excludes.Contains(child))
					continue;
				if (this.Excludes != null && this.ExcludeTypes.Contains(child.GetType()))
					continue;

				// ScrollableControlにイベントをバインドする
				if (child is ScrollableControl)
				{
					// イベントをバインドする
					if (bind)
					{
						child.MouseDown += this.ChildControl_MouseDown;
						child.MouseUp += this.ChildControl_MouseUp;
						child.MouseMove += this.ChildControl_MouseMove;
					}
					// アンバインドする
					else
					{
						child.MouseDown -= this.ChildControl_MouseDown;
						child.MouseUp -= this.ChildControl_MouseUp;
						child.MouseMove -= this.ChildControl_MouseMove;
					}
				}
				else
				{
					// イベントをバインドする
					if (bind)
					{
						// ディクショナリに追加する
						if (!this.ControlDictionary.ContainsKey(child.Handle))
							this.ControlDictionary.Add(child.Handle, child);

						WndProcHooker.HookWndProc(child, new WndProcHooker.WndProcCallback(this.WM_LButtonDown_Handler), Win32.WM_LBUTTONDOWN);
						WndProcHooker.HookWndProc(child, new WndProcHooker.WndProcCallback(this.WM_LBUTTONUP_Handler), Win32.WM_LBUTTONUP);
						WndProcHooker.HookWndProc(child, new WndProcHooker.WndProcCallback(this.WM_MOUSEMOVE_Handler), Win32.WM_MOUSEMOVE);
					}
					// アンバインドする
					else
					{
						// ディクショナリから削除する
						if (this.ControlDictionary.ContainsKey(child.Handle))
							this.ControlDictionary.Remove(child.Handle);

						WndProcHooker.UnhookWndProc(child, Win32.WM_LBUTTONDOWN);
						WndProcHooker.UnhookWndProc(child, Win32.WM_LBUTTONUP);
						WndProcHooker.UnhookWndProc(child, Win32.WM_MOUSEMOVE);
					}

				}

				// この子コントロールにイベントをバインドする
				this.EventBind(child, bind);
			}
		}

		/// <summary>
		/// スクロール対象コントロールをスクロールさせます。
		/// </summary>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <returns>スクロール先の座標を取得します。</returns>
		private void Scroll(int x, int y)
		{
			// スクロール位置を取得する
			int moveX = (this.MouseDownScrollPosition.X * -1) + this.MouseDownLocation.X - x;
			int moveY = (this.MouseDownScrollPosition.Y * -1) + this.MouseDownLocation.Y - y;

			// 移動距離がない→戻る
			if (moveX == 0 && moveY == 0)
				return;
			
			// スクロールさせる
			this.Target.AutoScrollPosition = new Point(moveX, moveY);

			// 返す
			return;
		}

		#endregion

		#region イベントメソッド(スクロール用)

		int WM_LButtonDown_Handler(IntPtr hwnd, uint msg, uint wParam, int lParam, ref bool handled)
		{
			// 送信元コントロールがない→戻る
			if (!this.ControlDictionary.ContainsKey(hwnd))
				return 0;

			// 送信元コントロールを取得する
			Control control = this.ControlDictionary[hwnd];

			// マウスのクライアント座標を取得する
			int x = lParam & 0xffff;
			int y = lParam >> 16;

			// イベントを発生させる
			this.ChildControl_MouseDown(control, new MouseEventArgs(MouseButtons.Left, 1, x, y, 0));

			// ハンドリングする
		//	handled = true;

			//System.Diagnostics.Debug.WriteLine(string.Format("WM_LBUTTONDOWN : Position {0}x{1} : Handled {2} : Name {3}", x, y, handled, control.Text));

			// 戻る
			return 0;
		}

		int WM_LBUTTONUP_Handler(IntPtr hwnd, uint msg, uint wParam, int lParam, ref bool handled)
		{
			// 送信元コントロールがない→戻る
			if (!this.ControlDictionary.ContainsKey(hwnd))
				return 0;

			// 送信元コントロールを取得する
			Control control = this.ControlDictionary[hwnd];

			// マウスのクライアント座標を取得する
			int x = lParam & 0xffff;
			int y = lParam >> 16;

			// イベントを発生させる
			this.ChildControl_MouseUp(control, new MouseEventArgs(MouseButtons.Left, 1, x, y, 0));

			// ハンドリングする
		//	handled = true;

			//System.Diagnostics.Debug.WriteLine(string.Format("WM_LBUTTONUP   : Position {0}x{1} : Handled {2} : Name {3}", x, y, handled, control.Text));

			// 戻る
			return 0;
		}

		int WM_MOUSEMOVE_Handler(IntPtr hwnd, uint msg, uint wParam, int lParam, ref bool handled)
		{
			// 送信元コントロールがない→戻る
			if (!this.ControlDictionary.ContainsKey(hwnd))
				return 0;

			// 送信元コントロールを取得する
			Control control = this.ControlDictionary[hwnd];

			// マウスのクライアント座標を取得する
			int x = lParam & 0xffff;
			int y = lParam >> 16;

			// イベントを発生させる
			this.ChildControl_MouseMove(control, new MouseEventArgs(MouseButtons.Left, 1, x, y, 0));

			// ハンドリングする
		//	handled = true;

			//System.Diagnostics.Debug.WriteLine(string.Format("WM_MOUSEMOVE   : Position {0}x{1} : Handled {2} : Name {3}", x, y, handled, control.Text));

			// 戻る
			return 0;
		}

		private void ChildControl_MouseDown(object sender, MouseEventArgs e)
		{
			// 呼び出し元コントロールを取得する
			Control control = (Control)sender;

			// マウスが押された時の各種情報を格納しておく
			this.NowMouseDown = true;

			// マウス座標をスクリーン座標にして格納しておく
			this.MouseDownLocation = control.PointToScreen(new Point(e.X, e.Y));

			// 現在のスクロール位置を格納しておく
			this.MouseDownScrollPosition = this.Target.AutoScrollPosition;
		}

		private void ChildControl_MouseUp(object sender, MouseEventArgs e)
		{
			// マウスが押されたフラグをオフにする
			this.NowMouseDown = false;
		}

		private void ChildControl_MouseMove(object sender, MouseEventArgs e)
		{
			// マウスが押されてなければ戻る
			if (!this.NowMouseDown)
				return;

			// 呼び出し元コントロールを取得する
			Control control = (Control)sender;

			// マウス座標をスクリーン座標に変換する
			Point point = control.PointToScreen(new Point(e.X, e.Y));
			
			// スクロールする
			this.Scroll(point.X, point.Y);
		}

		#endregion

		private bool IsOverWM65()
		{
			// バージョンを取得する
			Version osVer = System.Environment.OSVersion.Version;

			// WM 6.5 以降であるかどうか判定
			if (osVer.Major >= 5 && osVer.Minor >= 2)
			{
				// OS 6.5 以降である
				if (osVer.Build > 21000)
				{
					return true;
				}
			}

			// WM6.5 以降ではない
			return false;
		}

	}
}
